from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from enum import Enum
import re
import os

from autohandshake.src.exceptions import (
    InvalidURLError, NoSuchElementError, WrongPageForMethodError,
    InsufficientPermissionsError, InvalidUserTypeError
)
from autohandshake.src.constants import MAX_WAIT_TIME
from autohandshake.src.constants import BASE_URL


class UserType(Enum):
    """
    The possible user types in Handshake

            * Employer - a Handshake employer account
            * Staff - a career services staff/admin account
            * Student - a student or alumni account
    """

    EMPLOYER = 'Employers'
    STAFF = 'Career Services'
    STUDENT = 'Students'


class HandshakeBrowser:
    """
    An automated browser for navigating Handshake.

    Since a logged-in instance of this class is returned by HandshakeSession's
    __enter__ method, it does not usually need to be manually instantiated. Additionally,
    for most use cases, the user only needs to pass a HandshakeBrowser object to
    a Page object, then let the Page's methods do the heavy-lifting.

    For example, you almost never need to write:
    ::
        browser = HandshakeBrowser()

    The vast majority of use cases look something like:
    ::
        with HandshakeSession(school_url, email) as browser:
            some_page = SomePage(browser)
            some_page.do_something()

    If you need to specify a custom max_wait_time, that can be done through the
    HandshakeSession object:
    ::
        # this
        with HandshakeSession(school_url, email, max_wait_time = 60) as browser:
            some_page = SomePage(browser)

        # not this
        browser = HandshakeBrowser(max_wait_time = 60)

    """

    def __init__(self, max_wait_time: int = MAX_WAIT_TIME):
        """
        :param max_wait_time: the maximum time (in seconds) to wait for an element
                              to load before throwing a timeout error
        :type max_wait_time: int
        """
        options = webdriver.ChromeOptions()
        options.add_argument('--window-size=1920,1080')

        dirname = os.path.dirname(__file__)
        driver_path = os.path.join(dirname, '../chromedriver.exe')

        self._browser = webdriver.Chrome(executable_path=driver_path,
                                         options=options)
        self.max_wait_time = max_wait_time
        self._user_type = None

    def get(self, url: str):
        """Go to the web page specified by the given Handshake url.

        :param url: the url to visit. Must be of the form
                    "https://[...].joinhandshake.com[/...]"
        :type url: str
        """
        self._validate_url_str(url)
        self._browser.get(url)
        self._validate_page_exists()
        self._validate_permissions()

    def quit(self):
        """Close the browser"""
        self._browser.quit()

    def element_exists_by_xpath(self, xpath: str) -> bool:
        """
        Determine whether or not an element with the given xpath exists in the page.

        :param xpath: the xpath of the element to search for
        :type xpath: str
        :return: True if the element exists, false otherwise
        :rtype: bool
        """
        try:
            self._browser.find_element_by_xpath(xpath)
            return True
        except NoSuchElementException:
            return False

    def wait_until_element_exists_by_xpath(self, xpath: str):
        """
        Wait until an element with the given xpath exists on the page.

        :param xpath: the xpath of the element to wait for
        :type xpath: str
        """
        try:
            WebDriverWait(self._browser, self.max_wait_time).until(
                EC.visibility_of_element_located((By.XPATH, xpath)))
        except TimeoutException:
            raise TimeoutError(f"Element with xpath {xpath} did not appear in time")

    def wait_until_element_does_not_exist_by_xpath(self, xpath: str):
        """
        Wait until an element with the given xpath exists on the page.

        :param xpath: the xpath of the element to wait for
        :type xpath: str
        """
        try:
            WebDriverWait(self._browser, self.max_wait_time).until(
                EC.invisibility_of_element_located((By.XPATH, xpath)))
        except TimeoutException:
            raise TimeoutError(f"Element with xpath {xpath} did not disappear in tme")

    def wait_until_element_is_clickable_by_xpath(self, xpath: str):
        """
        Wait until an element with the given xpath is clickable.

        :param xpath: the xpath of the element to wait for
        :type xpath: str
        """
        try:
            WebDriverWait(self._browser, self.max_wait_time).until(
                EC.element_to_be_clickable((By.XPATH, xpath)))
        except TimeoutException:
            raise TimeoutError(f"Element with xpath {xpath} did not become clickable")

    def send_text_to_element_by_xpath(self, xpath: str, text: str, clear: bool = True):
        """
        Send a string to an input field identified by the given xpath

        :param text: the text to send
        :type text: str
        :param xpath: the xpath of the input field to which to send the text
        :type xpath: str
        :param clear: whether or not to clear the field before sending text. If
                      False, text will be appended to any text already present.
        :type clear: bool
        """
        try:
            element = self._browser.find_element_by_xpath(xpath)
            if clear:
                element.clear()
            element.send_keys(text)
        except NoSuchElementException:
            raise NoSuchElementError(f'No element found for xpath: "{xpath}"')

    def click_element_by_xpath(self, xpath):
        """
        Click an element on the page given its xpath

        :param xpath: the xpath of the element to click
        :type xpath: str
        """
        try:
            self._browser.find_element_by_xpath(xpath).click()
        except NoSuchElementException:
            raise NoSuchElementError(f'No element found for xpath: "{xpath}"')

    def wait_then_click_element_by_xpath(self, xpath):
        """
        Click an element on the page given its xpath after waiting to make sure
        it exists

        :param xpath: the xpath of the element to click
        :type xpath: str
        """
        self.wait_until_element_exists_by_xpath(xpath)
        self.click_element_by_xpath(xpath)

    def get_element_attribute_by_xpath(self, xpath: str, attribute: str) -> str:
        """
        Get the value of the given attribute from the element with the given xpath

        :param xpath: the xpath of the element of interest
        :type xpath: str
        :param attribute: the name of the attribute of interest, e.g. 'value'
        :type attribute: str
        :return: the value of the attribute on the element of interest
        :rtype: str
        """
        try:
            if attribute.lower() == 'text':
                return self._browser.find_element_by_xpath(xpath).text
            return self._browser.find_element_by_xpath(xpath).get_attribute(attribute)
        except NoSuchElementException:
            raise NoSuchElementError(f'No element found for xpath: "{xpath}"')

    def get_elements_attribute_by_xpath(self, xpath: str, attribute: str) -> list:
        """
        Get the value of a given attribute for all elements with the given xpath

        :param xpath: the xpath of the elements of interest
        :type xpath: str
        :param attribute: the name of the attribute of interest, e.g. 'value'
        :type attribute: str
        :return: a list of values of the given attribute for each matching element
        :rtype: list
        """
        try:
            elements = self._browser.find_elements_by_xpath(xpath)
            if attribute.lower() == 'text':
                return [element.text for element in elements]
            return [element.get_attribute(attribute) for element in elements]
        except NoSuchElementException:
            raise NoSuchElementError(f'No elements found for xpath: "{xpath}"')

    def execute_script_on_element_by_xpath(self, script: str, xpath: str = None):
        """
        Execute the given javascript expression. If xpath of element is provided,
        the element becomes available to use in the script, and can be accessed
        using arguments[0].

        :param script: the javascript to be executed
        :type script: str
        :param xpath: the xpath of the optional element to be passed to the script
        :type xpath: str
        """
        if not xpath:
            self._browser.execute_script(script)
        else:
            try:
                self._browser.execute_script(script, self._browser.find_element_by_xpath(xpath))
            except NoSuchElementException:
                raise NoSuchElementError(f'No elements found for xpath: "{xpath}"')

    def element_is_selected_by_xpath(self, xpath: str) -> bool:
        """Get whether or not the element specified by the given xpath is selected

        :param xpath: the xpath of the elements of interest
        :type xpath: str
        :return: True if the element is selected, False otherwise
        :rtype: bool
        """
        try:
            return self._browser.find_element_by_xpath(xpath).is_selected()
        except:
            raise NoSuchElementError(f'No elements found for xpath: "{xpath}"')

    def switch_to_frame_by_xpath(self, xpath):
        try:
            frame = self._browser.find_element_by_xpath(xpath)
            self._browser.switch_to.frame(frame)
        except NoSuchElementException:
            raise NoSuchElementError(f'No elements found for xpath: "{xpath}"')

    def update_constants(self):
        """
        Update any Handshake environment constants such as School ID or User ID.

        This should be done every time the browser switches user types and upon
        initial login.
        """
        if self._user_type == UserType.EMPLOYER:
            self.employer_id = self._get_meta_constant('logged_in_user_institution_id')
        else:
            self.school_id = self._get_meta_constant('logged_in_user_institution_id')
        self.user_id = self._get_meta_constant('logged_in_user_id')
        self._user_type = UserType(self._get_meta_constant('current_user_type'))

    def _get_meta_constant(self, name: str) -> str:
        """
        Get the content of a meta tag with the given name.

        The method is used to pull data like the current school id, user id,
        or employer id from Handshake's <meta> tags. All tags are of the form:
        <meta content="foo" name="bar">. Given "bar," this method returns "foo".

        :param name: the name of the meta tag to query
        :type name: str
        :return: the content of the meta tag
        :rtype: str
        """
        return self.get_element_attribute_by_xpath(f'//meta[@name="{name}"]',
                                                   'content')

    def switch_to_new_tab(self):
        """Wait for the new tab to finish loaded, then switch to it."""
        WebDriverWait(self._browser, self.max_wait_time).until(
            EC.number_of_windows_to_be(2))
        self._browser.switch_to.window(self._browser.window_handles[-1])

    def return_to_main_tab(self):
        """With a second tab open, close the current tab and return to the main tab."""
        self._browser.execute_script('window.close();')
        self._browser.switch_to.window(self._browser.window_handles[0])

    def maximize_window(self):
        """Maximize the browser window."""
        self._browser.maximize_window()

    @property
    def user_type(self) -> UserType:
        """
        Get the user type of the account currently logged into Handshake.

        :return: the browser's currently-logged-in user type
        :rtype: UserType
        """
        return self._user_type

    def switch_users(self, user_type: UserType):
        """
        Switch to the system view specified by the given user type.

        This method automates the built-in "Switch Users" function in Handshake.

        :param user_type: the user type to which to switch
        :type user_type: UserType
        """
        if self._user_type == user_type:
            return  # do nothing if the browser already has the desired user type
        self.get(f'{BASE_URL}/user_switcher/options')
        try:
            self.click_element_by_xpath(f'//a[div[h4[contains(text(), "{user_type.value}")]]]')
        except NoSuchElementError:
            raise InvalidUserTypeError("User does not have a linked account of the given type")
        self.update_constants()

    @property
    def current_url(self):
        """Get the url of the browser's current page"""
        return self._browser.current_url

    @staticmethod
    def _validate_url_str(url: str):
        """
        Determine whether or not a given Handshake URL is valid

        :param login_url: the login url to test
        :type login_url: str
        """
        try:
            re.match(r'^https://[a-zA-Z]+\.joinhandshake\.com', url) \
                .group(0)
        except AttributeError:
            raise InvalidURLError('URL must be of the form '
                                  '"https://app.joinhandshake.com[/...]" or '
                                  '"https://[school name].joinhandshake.com[/...]"')

    def _validate_page_exists(self):
        """Determine whether the current page exists or gave a 404 error."""
        if self.element_exists_by_xpath("//p[contains(text(), 'You may want "
                                        "to head back to the homepage.')]"):
            raise InvalidURLError(self._browser.current_url)

    def _validate_permissions(self):
        """Determine whether or not the logged in user has permission to view the current page"""
        if self.element_exists_by_xpath("//div[contains(text(), 'You do not "
                                        "have permission to do that.')]"):
            raise InsufficientPermissionsError



from typing import List

from autohandshake.src.HandshakeBrowser import HandshakeBrowser, UserType
from autohandshake.src.Pages import Page
from autohandshake.src.constants import BASE_URL
from autohandshake.src.exceptions import NoSuchElementError


class InterviewSchedulePage(Page):
    """The main page for an Interview Schedule."""

    def __init__(self, schedule_id: int, browser: HandshakeBrowser):
        """
        :param type_id: the id of the interview schedule to load
        :type type_id: int
        :param browser: a logged-in HandshakeBrowser
        :type browser: HandshakeBrowser
        """
        self._id = schedule_id
        super().__init__(f'{BASE_URL}/interview_schedules/{schedule_id}', browser)

    @Page.require_user_type(UserType.STAFF)
    def get_contact_info(self) -> List[dict]:
        """
        Retrieve the name and email of each person listed under "Contacts" on the interview schedule page.

        Method returns a list of dicts of the form:
        ::

            [
                {'name': 'Jane Recruiter', 'email': 'jrec@company.com'},
                {'name': 'John Employer', 'email': 'jemp@company.com'}
            ]

        :return: a list of dicts containing contact info
        """

        contacts = []
        there_are_unprocessed_contacts = True
        i = 1
        email_xpath = '//p[contains(@data-bind, "text: email_address")]'
        close_btn_xpath = '//div[@id="show-contact"]//div[@class="modal-header"]//button[@data-dismiss="modal"]'

        while there_are_unprocessed_contacts:
            contact_link_xpath = f'//p[@data-bind="foreach: contacts"]/a[{i}]'
            try:
                # get contact name
                name = self._browser.get_element_attribute_by_xpath(contact_link_xpath, 'text')
                # open modal and get contact email
                self._browser.click_element_by_xpath(contact_link_xpath)
                self._browser.wait_until_element_exists_by_xpath(email_xpath)
                email = self._browser.get_element_attribute_by_xpath(email_xpath, 'text')
                contacts.append({'name': name, 'email': email})
                # close modal and move onto next contact
                self._browser.wait_until_element_is_clickable_by_xpath(close_btn_xpath)
                self._browser.click_element_by_xpath(close_btn_xpath)
                self._browser.wait_until_element_does_not_exist_by_xpath(
                    '//*[@id="show-contact"]//div[@class="modal-content"]')
                i += 1
            except NoSuchElementError:
                there_are_unprocessed_contacts = False
        return contacts

    @Page.require_user_type(UserType.STAFF)
    def get_reserved_rooms(self):
        """
        Get the number of rooms reserved for the interview schedule.

        :return: the number of rooms reserved for the schedule
        """
        reserved_text = self._browser.get_element_attribute_by_xpath('//span[contains(text(), " reserved")]', 'text')
        return int(reserved_text.split()[0])


    def _validate_url(self, url):
        """
        Ensure that the given URL is a valid URL.

        Since the URL is not entered by the user, it is always valid.

        :param url: the url to validate
        :type url: str
        """
        return

    def _wait_until_page_is_loaded(self):
        """Wait until the page has finished loading."""
        self._browser.wait_until_element_exists_by_xpath('//*[@class="property-box"]/h4[text()="Employer"]')



from autohandshake.src.exceptions import *
from autohandshake.src.HandshakeBrowser import HandshakeBrowser, UserType
from autohandshake.src.HandshakeSession import HandshakeSession
from autohandshake.src.Pages import AccessRequestPage, InsightsPage, \
    MajorSettingsPage, AppointmentCalendarPage, AppointmentTypePage, \
    AppointmentTypesListPage, InterviewSchedulePage, FileType, RequestStatus



import unittest
from autohandshake.src.Pages import AccessRequestPage, RequestStatus
from autohandshake.tests import TestSession
from datetime import datetime


class TestAccessRequestPage(unittest.TestCase):

    def test_get_request_data_with_single_page_data(self):
        with TestSession() as browser:
            expected_failed = [
                {
                    'user': 'Diunase Nso Keshu',
                    'user_id': '16052363',
                    'email': 'dnsokes1@jh.edu',
                    'request_date': datetime.strptime('2018-09-20', '%Y-%m-%d').date(),
                    'request': 'Student Access',
                    'status': 'failed'
                },
                {
                    'user': 'Songnan Wang',
                    'user_id': '374064',
                    'email': 'swang140@jhu.edu',
                    'request_date': datetime.strptime('2018-09-05', '%Y-%m-%d').date(),
                    'request': 'Student Reactivation',
                    'status': 'failed'
                },
                {
                    'user': 'Kristina Grubb',
                    'user_id': '369508',
                    'email': 'kgrubb4@jhu.edu',
                    'request_date': datetime.strptime('2018-08-15', '%Y-%m-%d').date(),
                    'request': 'Student Reactivation',
                    'status': 'failed'
                }
            ]
            access_page = AccessRequestPage(browser)
            actual_failed = access_page.get_request_data(RequestStatus.FAILED)
            self.assertEqual(expected_failed, actual_failed[-3:])

    # This test should only be uncomment and run manually to visually confirm the
    # code works. The actual results of this method call vary from day to day, and
    # a true, consistent test of this functionality is not possible.

    # def test_get_request_data_with_multi_page_data(self):
    #
    #     with TestSession(10) as browser:
    #         access_page = AccessRequestPage(browser)
    #
    #         print(access_page.get_request_data())



from autohandshake.src.Pages.LoginPage import LoginPage
from autohandshake.src.Pages.Page import Page
from autohandshake.src.Pages.InsightsPage import InsightsPage, FileType
from autohandshake.src.Pages.AccessRequestPage import AccessRequestPage, RequestStatus
from autohandshake.src.Pages.AppointmentCalendarPage import AppointmentCalendarPage
from autohandshake.src.Pages.InterviewSchedulePage import InterviewSchedulePage

from autohandshake.src.Pages.SchoolSettings import *



from autohandshake.src.Pages import Page
from autohandshake.src.exceptions import NoSuchElementError
from autohandshake.src.HandshakeBrowser import HandshakeBrowser, UserType
from autohandshake.src.constants import BASE_URL
from dateutil.parser import parse
from bs4 import BeautifulSoup
from typing import List, Generator
import datetime
import re


class AppointmentCalendarPage(Page):
    """
    The calendar view of staff appointments.

    Can only be accessed with a career services/admin-type account.
    """

    def __init__(self, browser: HandshakeBrowser):
        """
        :param browser: a logged-in HandshakeBrowser
        :type browser: HandshakeBrowser
        """
        super().__init__(f'{BASE_URL}/appointments/calendar', browser)

    @Page.require_user_type(UserType.STAFF)
    def go_to_date(self, date: datetime.date):
        """Using the calendar's datepicker, navigate to the specified date

        :param date: the date to which to navigate on the calendar
        :type date: date
        """
        OPEN_DATEPICKER_XPATH = "//span[@id='date-button']"
        DATEPICKER_CALENDAR_XPATH = "//div[@class='datetimepicker datetimepicker-inline']"
        DAYS_TO_MONTHS_SWITCH_XPATH = "//div[@class='datetimepicker-days']//th[@class='switch']"
        MONTHS_TO_YEARS_SWITCH_XPATH = "//div[@class='datetimepicker-months']//th[@class='switch']"
        YEAR_XPATH = f"//span[text()='{date.year}']"
        MONTH_XPATH = f"//span[text()='{date.strftime('%b')}']"
        DAY_XPATH = f"//td[text()='{date.day}' and @class='day']"
        LOADING_SPINNER_XPATH = '//span[@class="ladda-spinner"]/div'

        # open datepicker and prepare to select date
        self._browser.wait_then_click_element_by_xpath(OPEN_DATEPICKER_XPATH)
        self._browser.wait_until_element_exists_by_xpath(DATEPICKER_CALENDAR_XPATH)
        self._browser.wait_then_click_element_by_xpath(DAYS_TO_MONTHS_SWITCH_XPATH)
        self._browser.wait_then_click_element_by_xpath(MONTHS_TO_YEARS_SWITCH_XPATH)

        # select year, month, and day, in that order
        self._browser.click_element_by_xpath(YEAR_XPATH)
        self._browser.click_element_by_xpath(MONTH_XPATH)
        self._browser.click_element_by_xpath(DAY_XPATH)

        # wait for new calendar date to finish loading
        self._browser.wait_until_element_does_not_exist_by_xpath(LOADING_SPINNER_XPATH)

    @Page.require_user_type(UserType.STAFF)
    def get_unfilled_blocks(self, start_date: datetime.date, end_date: datetime.date = None,
                            include_mediums: bool = False) -> List[dict]:
        """
        Get a list of unfilled appointment block data for the days from
        start_date up to (but not including) end_date.

        The method returns a list of appointment block dicts of the form:
        ::

            {
                'start_time': datetime.datetime(2019, 3, 7, 9, 30), # block start time
                'end_time': datetime.datetime(2019, 3, 7, 10, 30), # block end time
                'length': 60, # difference in minutes between start time and end time
                'status': 'unfilled', # status of the appointment
                'staff_name': 'Jane Coach', # the name of the staff member associated with the block
                'mediums': ['in-person', 'virtual'] # [OPTIONAL] the list of possible mediums for the open block
            }

        :param start_date: the start date of the date range from which to pull the data
        :type start_date: date
        :param end_date: the end date of the date range from which to pull the data.
                         If no end date is specified, only pull data for start_date
        :type end_date: date
        :param include_mediums: whether or not to include the appointment block mediums
                                in the data. Including the medium significantly
                                increases the time required to pull the data.
        :type include_mediums: bool
        :return: a list of appointment block dicts
        :rtype: list
        """
        if end_date:
            results = []
            for date in self._daterange(start_date, end_date):
                results += self._get_unfilled_on_one_date(date, include_mediums=include_mediums)
        else:
            results = self._get_unfilled_on_one_date(start_date, include_mediums=include_mediums)

        return results

    def _validate_url(self, url):
        """
        Ensure that the given URL is a valid URL.

        Since the URL is not entered by the user, it is always valid.

        :param url: the url to validate
        :type url: str
        """
        return

    def _wait_until_page_is_loaded(self):
        """Wait until the page has finished loading."""

        # altering the window size is necessary because the calendar's responsive
        # design only triggers upon resizing, not on initial load
        self._browser.maximize_window()
        self._browser.wait_until_element_exists_by_xpath("//div[@class='appointment-calendar-staff-container']")

    def _get_unfilled_on_one_date(self, date: datetime.date, include_mediums: bool = False) -> List[dict]:
        """
        Get unfilled block data for a single date.

        :param date: the date for which to get data
        :type date: date
        :param include_mediums: bool
        :type include_mediums: whether or not to include appointment mediums in the data
        :return: a list of unfilled appointment block data dicts
        :rtype: list
        """

        # the height of a calendar block increases by 3 pixels for each additional
        # minute of length
        PIXELS_PER_MINUTE = 3
        DATE_STR = date.strftime("%m/%d/%y")

        self.go_to_date(date)

        staff_header = BeautifulSoup(self._browser.get_element_attribute_by_xpath(
            "//div[@class='appointment-calendar-staff-container']", "innerHTML"),
            'html.parser')

        appt_grid = BeautifulSoup(self._browser.get_element_attribute_by_xpath(
            "//div[@class='appointment-calendar-grid-container']", "innerHTML"),
            'html.parser')

        # Populate ordered list with staff as they appear in the header
        staff = []
        for link in staff_header.findAll('a'):
            staff.append({
                'name': link.text,
                'blocks': {
                    'available': [],
                    'appointments': [],
                    'blocked': [],
                }
            })

        # for each staff column in the calendar...
        for column_index, column in enumerate(appt_grid.findAll(
                'div', 'appointment-calendar-advisor-data')):

            # capture all available, appointment, and unavailable blocks
            available_blocks = column.findAll(lambda x:
                                              x.name == 'div' and
                                              'appointment-block' in x.get('class', []) and
                                              not 'unavailable' in x['class'])

            appointment_blocks = column.findAll('div', 'appointment')
            unavailable_blocks = column.findAll('div', 'unavailable')

            # Capture "Available" blocks
            for block_index, block in enumerate(available_blocks):
                block_start_time = parse(block['data-time'][:-6])
                block_height = int(re.search("height: (\d+)px", block['style']).group(1))
                block_length = block_height / PIXELS_PER_MINUTE
                block_end_time = block_start_time + datetime.timedelta(minutes=block_length)
                staff[column_index]['blocks']['available'].append({
                    'start_time': block_start_time,
                    'end_time': block_end_time,
                    'x_idx': column_index + 1,
                    'y_idx': block_index + 1
                })

            # Capture "Appointment" blocks
            for appt_block in appointment_blocks:
                appt_time_text = appt_block.find('div', 'timing-information').text
                appt_start_time = parse(DATE_STR + ' ' + re.search('^(\d+:\d+ [a-z]+)', appt_time_text).group(1))
                appt_end_time = parse(DATE_STR + ' ' + re.search('(\d+:\d+ [a-z]+)$', appt_time_text).group(1))

                staff[column_index]['blocks']['appointments'].append({
                    'start_time': appt_start_time,
                    'end_time': appt_end_time,
                })

            # Capture "Unavailable" blocks
            for unavail_block in unavailable_blocks:
                block_start_time = parse(unavail_block['data-time'][:-6])
                block_height = int(re.search("height: (\d+)px", unavail_block['style']).group(1))
                block_length = block_height / PIXELS_PER_MINUTE
                block_end_time = block_start_time + datetime.timedelta(minutes=block_length)

                staff[column_index]['blocks']['blocked'].append({
                    'start_time': block_start_time,
                    'end_time': block_end_time
                })

        # cross reference available, unavailable, and appointment blocks to
        # determine unfilled time ranges
        unfilled_appt_list = []

        for staff_member in staff:
            all_conflicts = staff_member['blocks']['appointments'] + staff_member['blocks']['blocked']
            all_open = staff_member['blocks']['available'][:]
            for open_block in all_open:

                # break each open block up into its constituent open segments that are not covered by unavailable or appt blocks
                open_segments = self._get_open_block_segments(open_block, all_conflicts)
                for open_segment in open_segments:
                    unfilled_block = {
                        'start_time': open_segment['start_time'],
                        'end_time': open_segment['end_time'],
                        'length': int((open_segment['end_time'] - open_segment['start_time']).total_seconds() / 60),
                        'status': 'unfilled',
                        'staff_name': staff_member['name'],
                    }
                    if include_mediums:
                        unfilled_block['mediums'] = self._get_appt_mediums(open_block['x_idx'],
                                                                           open_block['y_idx'])
                    unfilled_appt_list.append(unfilled_block)

        return unfilled_appt_list

    @staticmethod
    def _get_open_block_segments(open_block: dict, conflict_blocks: list) -> List[dict]:
        """
        Compare the given block with all possible conflict blocks and return the
        truly open segments of the original block.

        For example, if the open block stretches from 10 - 12, and there are conflicts
        from 9:50 - 10:10, 10:30 - 11:00, and 11:30-12:30, then this method would
        return a list of two open blocks: from 10:10 to 10:30 and from 11:00 to 11:30.

        :param open_block: an available calendar block
        :type open_block: dict
        :param conflict_blocks: a list of all unavailable calendar blocks
        :type conflict_blocks: list
        :return: a list of all remaining available segments of the open_block
        :rtype: list
        """
        open_segments = [open_block]
        for conflict in conflict_blocks:
            for segment in open_segments:
                # case where availability block is completely consumed by conflict block
                if conflict['start_time'] <= segment['start_time'] and \
                        conflict['end_time'] >= segment['end_time']:
                    open_segments.remove(segment)

                if conflict['start_time'] <= segment['start_time'] and \
                        conflict['end_time'] > segment['start_time'] and \
                        conflict['end_time'] < segment['end_time']:
                    segment['start_time'] = conflict['end_time']

                if conflict['start_time'] > segment['start_time'] and \
                        conflict['start_time'] < segment['end_time'] and \
                        conflict['end_time'] >= segment['end_time']:
                    segment['end_time'] = conflict['start_time']

                if conflict['start_time'] > segment['start_time'] and \
                        conflict['end_time'] < segment['end_time']:
                    open_segments.remove(segment)

                    first_half = dict(segment)
                    first_half['end_time'] = conflict['start_time']
                    open_segments.append(first_half)

                    second_half = dict(segment)
                    second_half['start_time'] = conflict['end_time']
                    second_half['end_time'] = segment['end_time']
                    open_segments.append(second_half)

        return open_segments

    def _get_appt_mediums(self, col_num: int, block_num: int) -> List[str]:
        """
        Get the appointment mediums for a single available block

        :param col_num: the block's calendar column number
        :type col_num: int
        :param block_num: the block's index within the column
        :type block_num: int
        :return: a list of the block's mediums
        :rtype: list
        """
        BLOCK_XPATH = f'//div[@class="appointment-calendar-advisor-data"]' \
            f'[{col_num}]/div[@class="appointment-block clickable"][{block_num}]'
        EDIT_ICON_XPATH = f"{BLOCK_XPATH}//*[contains(@class,'edit-icon')]"
        MEDIUMS_XPATH = '//div[@id="s2id_appointment-medium"]/ul/li/div'
        MODAL_XPATH = "//div[contains(@class, 'edit-appointment-block-popover')]"
        CLOSE_MODAL_XPATH = f'{MODAL_XPATH}//*[contains(@class, "btn-primary")]'
        LOADING_SPINNER_XPATH = '//span[@class="ladda-spinner"]/div'

        try:
            try:
                self._browser.click_element_by_xpath(EDIT_ICON_XPATH)
            except:
                # open modal using javascript to account for "element not clickable at (x, y)" error
                self._browser.execute_script_on_element_by_xpath('arguments[0].click()',
                                                                 EDIT_ICON_XPATH)
            self._browser.wait_until_element_exists_by_xpath(MODAL_XPATH)
            # capture the appointment mediums
            mediums = self._browser.get_elements_attribute_by_xpath(MEDIUMS_XPATH,
                                                                    'innerHTML')
        except NoSuchElementError:
            mediums = []

        # close the modal
        self._browser.wait_then_click_element_by_xpath(CLOSE_MODAL_XPATH)

        self._browser.wait_until_element_does_not_exist_by_xpath(
            MODAL_XPATH)
        self._browser.wait_until_element_does_not_exist_by_xpath(
            LOADING_SPINNER_XPATH)

        return mediums

    @staticmethod
    def _daterange(start_date: datetime.date,
                   end_date: datetime.date) -> Generator[datetime.date, None, None]:
        """
        Create a generator that yeilds dates from the start date up to (but not
        including) the end date.

        :param start_date: the start date of the range
        :type start_date: datetime.date
        :param end_date: the end date of the range
        :type end_date: datetime.date
        :return: the dates in between start_date and end_date
        :rtype: datetime.date
        """
        for n in range(int(((end_date - start_date).days))):
            yield start_date + datetime.timedelta(n)



from autohandshake.src.Pages import Page
from autohandshake.src.HandshakeBrowser import HandshakeBrowser, UserType
from autohandshake.src.constants import BASE_URL
from typing import List


class MajorSettingsPage(Page):

    def __init__(self, browser: HandshakeBrowser):
        """
        :param browser: a logged-in HandshakeBrowser
        :type browser: HandshakeBrowser
        """
        majors_base_url = f'{BASE_URL}/schools/{browser.school_id}/school_majors'
        url_query = '?ajax=true&query=&category=ManageSchoolMajors&page=1&per_page=5000&sort_direction=asc&sort_column=default&followed_only=false&qualified_only=&core_schools_only=false&including_all_facets_in_searches=false'
        super().__init__(majors_base_url + url_query,
                         browser)

    @Page.require_user_type(UserType.STAFF)
    def get_major_mapping(self) -> List[dict]:
        """
        Get a list of all the school's majors and their associated major groups.

        :return: a list of major-to-major-group mapping dicts, of the form:
                 {'major': 'major name', 'groups': ['group 1', 'group 2', etc]}
        :rtype: list
        """
        mappings = []

        majors = self._browser.get_elements_attribute_by_xpath("//tbody/tr/td[1]/a", "text")
        group_strings = self._browser.get_elements_attribute_by_xpath("//tbody/tr/td[2]/a", "text")

        for i in range(len(majors)):
            groups = self._parse_major_groups_string(group_strings[i])
            mappings.append({
                'major': majors[i],
                'groups': groups
            })

        return mappings

    def _validate_url(self, url):
        """
        Ensure that the given URL is a valid URL.

        Since the URL is not entered by the user, it is always valid.

        :param url: the url to validate
        :type url: str
        """
        return

    def _wait_until_page_is_loaded(self):
        """Wait until the page has finished loading."""
        self._browser.wait_until_element_exists_by_xpath("//tbody/tr")

    def _parse_major_groups_string(self, groups_string: str) -> List[str]:
        """
        Parse a comma-separated list of major groups into a list of distinct groups.

        :param groups_string: a comma-separated string of major groups
        :type groups_string: str
        :return: a list of the separate major groups
        :rtype: list
        """
        groups_with_commas_in_them = [
            'City, Urban, Regional Planning',
            'Parks, Recreation & Leisure Studies',
            'Radio, Television, Media'
        ]
        groups = groups_string.split(', ')
        for group in groups_with_commas_in_them:
            groups = self._correct_oversplit_string(groups, group)
        return groups

    @staticmethod
    def _correct_oversplit_string(str_list: List[str], joined_str: str) -> List[str]:
        """
        If every word in the given joined string is present in the given list,
        replace the separate words with a single instance of the joined string.

        Ex: you want to split the string "London, England, Boston" into the list
        ["London, England", "Boston"]. Merely splitting the string on commas will
        give you ["London", "England", "Boston"]. Calling this function with the
        arguments ["London", "England", "Boston"] and "London, England" will
        return the list ["London, England", "Boston"].

        :param str_list: a list of strings containing at least one overly-split
                         series of sub-strings
        :type str_list: list
        :param joined_str: the full string with which to replace the overly-split,
                           separate elements of this string in the list
        :type joined_str: str
        :return: the original list with all disparate elements of joined_str
                 replaced by the full joined_str
        :rtype: list
        """
        group_is_in_list = True
        group_words = joined_str.split(', ')
        for word in group_words:
            if not word in str_list:
                group_is_in_list = False

        if group_is_in_list:
            new_list = []
            for group_name in str_list:
                if not group_name in group_words:
                    new_list.append(group_name)
            new_list.append(joined_str)
            return new_list
        return str_list



from autohandshake.src.Pages import Page
from autohandshake.src.constants import BASE_URL
from autohandshake.src.HandshakeBrowser import HandshakeBrowser, UserType
from autohandshake.src.Pages.SchoolSettings.AppointmentTypePage import AppointmentTypePage
import re


class AppointmentTypesListPage(Page):
    """The overview settings page listing all appointment types"""

    def __init__(self, browser: HandshakeBrowser):
        """
        :param browser: a logged-in HandshakeBrowser
        :type browser: HandshakeBrowser
        """
        super().__init__(f'{BASE_URL}/schools/{browser.school_id}/appointment_types',
                         browser)

    @Page.require_user_type(UserType.STAFF)
    def get_type_ids(self) -> list:
        """Get a list of appointment type IDs for your school"""
        type_links = self._browser.get_elements_attribute_by_xpath(
            "//div[@class='top-right-btn']/a[@class='btn btn-default']", "href")
        return [self._extract_type_id_from_url(link) for link in type_links]

    @Page.require_user_type(UserType.STAFF)
    def get_type_settings(self, how_many: int = None) -> list:
        """
        Get a list of settings objects for all appointment type settings.

        The method returns a list of dicts, each of the form specified in :func:`autohandshake.AppointmentTypePage.get_settings()`

        :param how_many: if specified, the number of settings objects to collect.
                         If None (the default), collect all settings. This
                         argument is mainly for testing purposes, as downloading
                         the full list of settings is prohibitively slow for
                         testing.
        :type how_many: int
        :return: a list of type settings objects for all appointment types
        :rtype: list
        """
        settings_list = []
        type_ids = self.get_type_ids()
        if not how_many:
            how_many = len(type_ids)
        for i in range(0, how_many):
            id = type_ids[i]
            type_page = AppointmentTypePage(id, self._browser)
            settings_list.append(type_page.get_settings())
        return settings_list

    def _extract_type_id_from_url(self, url):
        """Given an appointment type settings url, extract the type id"""
        return int(re.search("/appointment_types/([0-9]+)/edit", url).group(1))

    def _validate_url(self, url):
        """
        Ensure that the given URL is a valid URL.

        Since the URL is not entered by the user, it is always valid.

        :param url: the url to validate
        :type url: str
        """
        return

    def _wait_until_page_is_loaded(self):
        """Wait until the page has finished loading."""
        self._browser.wait_until_element_exists_by_xpath("//*[@class='top-right-btn']")



import unittest

from autohandshake import HandshakeSession
from autohandshake.tests import email, password, homepage, school_id
from autohandshake.src.exceptions import InvalidPasswordError


class TestHandshakeSession(unittest.TestCase):

    def test_does_not_throw_error_given_valid_url(self):
        standard_url = homepage

        login_style_url = f"{homepage}/login"

        try:
            with HandshakeSession(login_url=standard_url, email=email, password=password):
                pass
        except ValueError:
            self.fail("Unexpected ValueError given a valid normal URL")

        try:
            with HandshakeSession(login_url=login_style_url, email=email, password=password):
                pass
        except ValueError:
            self.fail("Unexpected ValueError given a valid login-style URL")

    def test_sets_browser_school_id(self):
        try:
            with HandshakeSession(homepage, email, password) as browser:
                self.assertEqual(browser.school_id, school_id)
        except AttributeError:
            self.fail("School ID has not been set")

    def test_works_with_keyring_password(self):
        """This test requires a valid keyring password to be set up on the testing computer"""
        try:
            with HandshakeSession(homepage, email):
                pass
        except InvalidPasswordError:
            self.fail("Password did not work or was not found")

    def test_throws_error_if_keyring_cannot_find_password(self):
        """This test requires the testing computer to NOT have the specified service/user combination set"""
        with self.assertRaises(InvalidPasswordError):
            with HandshakeSession('https://abc.joinhandshake.com', 'notuser@abc.edu'):
                pass



from autohandshake.src.Pages.Page import Page
from autohandshake.src.HandshakeBrowser import HandshakeBrowser
from autohandshake.src.exceptions import InvalidURLError, NoSuchElementError, \
    InvalidEmailError, InvalidPasswordError
import re


class LoginPage(Page):
    """
    The old Handshake login page
    """

    def __init__(self, url: str, browser: HandshakeBrowser):
        """
        :param url: the url of the school's Handshake login page
        :type url: str
        :param browser: a HandshakeBrowser that has not logged in yet
        :type browser: HandshakeBrowser
        """
        super().__init__(url, browser)
        self.validate_url_school()

    def _wait_until_page_is_loaded(self):
        """Wait until the page has finished loading.

        Return immediately since there are no complex load conditions
        """
        return

    def _validate_url(self, url):
        """
        Ensure that the given URL is a valid login URL

        :param url: the url to validate
        :type url: str
        """
        try:
            re.match(r'^https://[a-zA-Z]+\.joinhandshake\.com(/login)?$', url) \
                .group(0)
        except AttributeError:
            raise InvalidURLError()

    def validate_url_school(self):
        """Ensure that the current URL leads to a valid school's login page"""
        if self._browser.element_exists_by_xpath('//span[text()=\'Please '
                                                 'select your school to '
                                                 'sign in.\']'):
            raise InvalidURLError("The school specified in the URL is not valid")

    @Page.require_user_type(None)  # requires the user to be logged out, i.e. no user type
    def login(self, email, password):
        """
        Log into Handshake using the given credentials

        :param email: the username with which to log in
        :type email: str
        :param password: the password with which to log in
        :type password: str
        """
        self._enter_email_address(email)
        self._enter_password(password)
        self._browser.wait_until_element_exists_by_xpath('//div[@class="Select-placeholder"]')

    def _enter_email_address(self, email):
        """Enter email address into input field"""
        EMAIL_INPUT_XPATH = "//input[@name='identifier']"

        try:  # if you get the old login page
            EMAIL_LINK_XPATH = "//div[@class='sign-with-email-address']//a"
            self._browser.click_element_by_xpath(EMAIL_LINK_XPATH)
            self._browser.send_text_to_element_by_xpath(EMAIL_INPUT_XPATH, email)
            EMAIL_BTN_XPATH = "//div[@class='login-main__email-box']/button"
            self._browser.click_element_by_xpath(EMAIL_BTN_XPATH)
            if self._browser.element_exists_by_xpath("//div[text()='Please enter a valid email address']"):
                raise InvalidEmailError(f"No account found for email {email}")

        except NoSuchElementError:  # if you get the new login page
            EMAIL_LINK_XPATH = "//div[@class='sign-in-with-email-address']//a"
            self._browser.click_element_by_xpath(EMAIL_LINK_XPATH)
            self._browser.send_text_to_element_by_xpath(EMAIL_INPUT_XPATH, email)
            EMAIL_BTN_XPATH = "//div[@class='actions']/button"
            self._browser.click_element_by_xpath(EMAIL_BTN_XPATH)
            if 'known_error_message_present=true' in self._browser.current_url:
                raise InvalidEmailError(f"No account found for email {email}")

    def _enter_password(self, password):
        """Enter password into input field after having successfully entered email"""
        try:  # if you get the old login page
            self._browser.click_element_by_xpath("//a[@class='no-underline']")
            self._browser.send_text_to_element_by_xpath("//input[@name='password']", password)
            self._browser.click_element_by_xpath("//input[@name='commit']")
            if self._browser.element_exists_by_xpath("//div[contains(text(), "
                                                     "'You entered an invalid password.')]"):
                raise InvalidPasswordError("Invalid password")
        except NoSuchElementError:  # if you get the new login page
            self._browser.click_element_by_xpath("//a[@class='alternate-login-link']")
            self._browser.send_text_to_element_by_xpath("//input[@name='password']", password)
            self._browser.click_element_by_xpath("//button")
            if self._browser.element_exists_by_xpath("//div[contains(text(), "
                                                     "'You entered an invalid password.')]"):
                raise InvalidPasswordError("Invalid password")



import unittest
from autohandshake import AppointmentCalendarPage
from autohandshake.tests import TestSession
from datetime import datetime

LOCATION_VARIES = 'In-Person (Location: Varies- Coach will contact you with location)'
IN_PERSON = 'In-Person (Location: Career Center--Garland Hall 389)'
VIRTUAL = 'Virtual Appointment (Coach will Contact You)'


class TestAppointmentCalendarPage(unittest.TestCase):

    def test_go_to_date(self):
        test_date_1 = datetime(2018, 9, 13).date()
        test_date_2 = datetime(2019, 10, 30).date()

        with TestSession() as browser:
            calendar = AppointmentCalendarPage(browser)
            calendar.go_to_date(test_date_1)
            self.assertTrue(browser.element_exists_by_xpath("//span[text()='Thursday, September 13th']"))
            calendar.go_to_date(test_date_2)
            self.assertTrue(browser.element_exists_by_xpath("//span[text()='Wednesday, October 30th']"))

    def test_get_one_day_of_unfilled(self):
        with TestSession() as browser:
            calendar = AppointmentCalendarPage(browser)
            expected = [
                {
                    'start_time': datetime(2017, 11, 8, 15, 0),
                    'end_time': datetime(2017, 11, 8, 16, 0),
                    'length': 60,
                    'status': 'unfilled',
                    'staff_name': 'Justin Lorts',
                    'mediums': [LOCATION_VARIES]
                },
                {
                    'start_time': datetime(2017, 11, 8, 14, 0),
                    'end_time': datetime(2017, 11, 8, 14, 30),
                    'length': 30,
                    'status': 'unfilled',
                    'staff_name': 'Nadine Goldberg',
                    'mediums': [IN_PERSON, VIRTUAL]
                },
                {
                    'start_time': datetime(2017, 11, 8, 11, 0),
                    'end_time': datetime(2017, 11, 8, 12, 0),
                    'length': 60,
                    'status': 'unfilled',
                    'staff_name': 'Sonjala Williams',
                    'mediums': [LOCATION_VARIES]
                },
                {
                    'start_time': datetime(2017, 11, 8, 13, 0),
                    'end_time': datetime(2017, 11, 8, 15, 0),
                    'length': 120,
                    'status': 'unfilled',
                    'staff_name': 'Sonjala Williams',
                    'mediums': [LOCATION_VARIES]
                }
            ]
            actual = calendar.get_unfilled_blocks(datetime(2017, 11, 8).date(), include_mediums=True)
            self.assertEqual(expected, actual)

    def test_multi_day(self):
        with TestSession(30) as browser:
            calendar = AppointmentCalendarPage(browser)
            expected = [
                {
                    'start_time': datetime(2017, 11, 10, 9, 30),
                    'end_time': datetime(2017, 11, 10, 10, 0),
                    'length': 30,
                    'status': 'unfilled',
                    'staff_name': 'Caroline Kelly',
                    'mediums': [VIRTUAL]
                },
                {
                    'start_time': datetime(2017, 11, 10, 11, 0),
                    'end_time': datetime(2017, 11, 10, 11, 30),
                    'length': 30,
                    'status': 'unfilled',
                    'staff_name': 'Caroline Kelly',
                    'mediums': [VIRTUAL]
                },
                {
                    'start_time': datetime(2017, 11, 10, 13, 0),
                    'end_time': datetime(2017, 11, 10, 14, 0),
                    'length': 60,
                    'status': 'unfilled',
                    'staff_name': 'Lauren Barrett',
                    'mediums': [VIRTUAL]
                },
                {
                    'start_time': datetime(2017, 11, 10, 10, 0),
                    'end_time': datetime(2017, 11, 10, 12, 0),
                    'length': 120,
                    'status': 'unfilled',
                    'staff_name': 'Sonjala Williams',
                    'mediums': [LOCATION_VARIES]
                },
                {
                    'start_time': datetime(2017, 11, 10, 14, 0),
                    'end_time': datetime(2017, 11, 10, 15, 30),
                    'length': 90,
                    'status': 'unfilled',
                    'staff_name': 'Sonjala Williams',
                    'mediums': [LOCATION_VARIES]
                },
                {
                    'start_time': datetime(2017, 11, 13, 9, 0),
                    'end_time': datetime(2017, 11, 13, 10, 0),
                    'length': 60,
                    'status': 'unfilled',
                    'staff_name': 'Andrea Wiseman',
                    'mediums': [IN_PERSON, VIRTUAL]
                },
                {
                    'start_time': datetime(2017, 11, 13, 10, 30),
                    'end_time': datetime(2017, 11, 13, 11, 30),
                    'length': 60,
                    'status': 'unfilled',
                    'staff_name': 'Caroline Kelly',
                    'mediums': [IN_PERSON, VIRTUAL]
                },
                {
                    'start_time': datetime(2017, 11, 13, 10, 30),
                    'end_time': datetime(2017, 11, 13, 12, 0),
                    'length': 90,
                    'status': 'unfilled',
                    'staff_name': 'Justin Lorts',
                    'mediums': [LOCATION_VARIES]
                },
                {
                    'start_time': datetime(2017, 11, 13, 14, 30),
                    'end_time': datetime(2017, 11, 13, 15, 0),
                    'length': 30,
                    'status': 'unfilled',
                    'staff_name': 'Lauren Barrett',
                    'mediums': [IN_PERSON, VIRTUAL]
                },
                {
                    'start_time': datetime(2017, 11, 13, 13, 0),
                    'end_time': datetime(2017, 11, 13, 16, 0),
                    'length': 180,
                    'status': 'unfilled',
                    'staff_name': 'Sonjala Williams',
                    'mediums': [LOCATION_VARIES]
                },
                {
                    'start_time': datetime(2017, 11, 13, 10, 0),
                    'end_time': datetime(2017, 11, 13, 11, 0),
                    'length': 60,
                    'status': 'unfilled',
                    'staff_name': 'Sonjala Williams',
                    'mediums': [LOCATION_VARIES]
                },
                {
                    'start_time': datetime(2017, 11, 13, 11, 30),
                    'end_time': datetime(2017, 11, 13, 12, 0),
                    'length': 30,
                    'status': 'unfilled',
                    'staff_name': 'Sonjala Williams',
                    'mediums': [LOCATION_VARIES]
                }
            ]
            actual = calendar.get_unfilled_blocks(datetime(2017, 11, 10).date(), datetime(2017, 11, 14).date(),
                                                  include_mediums=True)
            self.assertEqual(expected, actual)


class TestAppointmentCalendarPageOpenBlockAlgorithm(unittest.TestCase):
    def test_degernerate_cases(self):
        open_block = {
            'start_time': datetime(2018, 11, 2, 10, 0),
            'end_time': datetime(2018, 11, 2, 14, 0)
        }

        consuming_conflicts = [
            {
                'start_time': datetime(2018, 11, 2, 9, 30),
                'end_time': datetime(2018, 11, 2, 16, 10)
            }
        ]

        consuming_actual = AppointmentCalendarPage._get_open_block_segments(open_block, consuming_conflicts)
        no_conflict_actual = AppointmentCalendarPage._get_open_block_segments(open_block, [])
        self.assertEqual([], consuming_actual)
        self.assertEqual([open_block], no_conflict_actual)

    def test_bookends_only_case(self):
        open_block = {
            'start_time': datetime(2018, 11, 2, 10, 0),
            'end_time': datetime(2018, 11, 2, 14, 0)
        }

        bookends_conflicts = [
            {
                'start_time': datetime(2018, 11, 2, 9, 30),
                'end_time': datetime(2018, 11, 2, 10, 10)
            },
            {
                'start_time': datetime(2018, 11, 2, 13, 55),
                'end_time': datetime(2018, 11, 2, 14, 30)
            }
        ]

        expected = [
            {
                'start_time': datetime(2018, 11, 2, 10, 10),
                'end_time': datetime(2018, 11, 2, 13, 55)
            }
        ]

        actual = AppointmentCalendarPage._get_open_block_segments(open_block, bookends_conflicts)
        self.assertEqual(expected, actual)

    def test_bookends_with_middle_conflict_case(self):
        open_block = {
            'start_time': datetime(2018, 11, 2, 10, 0),
            'end_time': datetime(2018, 11, 2, 14, 0)
        }

        one_middle_conflicts = [
            {
                'start_time': datetime(2018, 11, 2, 8, 30),
                'end_time': datetime(2018, 11, 2, 9, 10)
            },
            {
                'start_time': datetime(2018, 11, 2, 9, 30),
                'end_time': datetime(2018, 11, 2, 10, 10)
            },
            {
                'start_time': datetime(2018, 11, 2, 11, 15),
                'end_time': datetime(2018, 11, 2, 12, 30)
            },
            {
                'start_time': datetime(2018, 11, 2, 13, 55),
                'end_time': datetime(2018, 11, 2, 14, 30)
            },
            {
                'start_time': datetime(2018, 11, 2, 14, 30),
                'end_time': datetime(2018, 11, 2, 15, 30)
            }
        ]

        expected = [
            {
                'start_time': datetime(2018, 11, 2, 10, 10),
                'end_time': datetime(2018, 11, 2, 11, 15)
            },
            {
                'start_time': datetime(2018, 11, 2, 12, 30),
                'end_time': datetime(2018, 11, 2, 13, 55)
            }
        ]

        actual = AppointmentCalendarPage._get_open_block_segments(open_block, one_middle_conflicts)
        self.assertEqual(expected, actual)

    def test_bookends_with_mutiple_middle_conflicts(self):
        open_block = {
            'start_time': datetime(2018, 11, 2, 10, 0),
            'end_time': datetime(2018, 11, 2, 14, 0)
        }

        one_middle_conflicts = [
            {
                'start_time': datetime(2018, 11, 2, 8, 30),
                'end_time': datetime(2018, 11, 2, 9, 10)
            },
            {
                'start_time': datetime(2018, 11, 2, 9, 30),
                'end_time': datetime(2018, 11, 2, 10, 10)
            },
            {
                'start_time': datetime(2018, 11, 2, 11, 15),
                'end_time': datetime(2018, 11, 2, 12, 30)
            },
            {
                'start_time': datetime(2018, 11, 2, 12, 45),
                'end_time': datetime(2018, 11, 2, 13, 0)
            },
            {
                'start_time': datetime(2018, 11, 2, 13, 30),
                'end_time': datetime(2018, 11, 2, 13, 50)
            }
        ]

        expected = [
            {
                'start_time': datetime(2018, 11, 2, 10, 10),
                'end_time': datetime(2018, 11, 2, 11, 15)
            },
            {
                'start_time': datetime(2018, 11, 2, 12, 30),
                'end_time': datetime(2018, 11, 2, 12, 45)
            },
            {
                'start_time': datetime(2018, 11, 2, 13, 00),
                'end_time': datetime(2018, 11, 2, 13, 30)
            },
            {
                'start_time': datetime(2018, 11, 2, 13, 50),
                'end_time': datetime(2018, 11, 2, 14, 00)
            }
        ]

        actual = AppointmentCalendarPage._get_open_block_segments(open_block, one_middle_conflicts)
        self.assertEqual(expected, actual)



import pathlib

from setuptools import setup, find_packages
from autohandshake import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.rst").read_text()

# This call to setup() does all the work
setup(
    name="autohandshake",
    version=__version__,
    description="A library for automating tasks on the Handshake career services platform",
    long_description=README,
    url="https://github.com/cedwards036/autohandshake",
    author="Christopher Edwards",
    author_email="cedwards036@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=['selenium', 'bs4']
)



from abc import ABC, abstractmethod
from autohandshake.src.HandshakeBrowser import HandshakeBrowser, UserType
from autohandshake.src.exceptions import InvalidURLError, WrongPageForMethodError, \
    InvalidUserTypeError
from typing import Callable


class Page(ABC):
    """
    A single web page on the Handshake site
    """

    def __init__(self, url: str, browser: HandshakeBrowser):
        """
        Create a page object for the given URL

        :param url: the url of the page to initialize
        :type url: str
        :param browser: a HandshakeBrowser that is logged in to Handshake
        :type browser: HandshakeBrowser
        """
        self._validate_url(url)
        self._url = url
        self._browser = browser
        self._browser.get(url)
        self._wait_until_page_is_loaded()

    @abstractmethod
    def _wait_until_page_is_loaded(self):
        """Wait until the page has finished loading.

        For pages without complex javascript involved in the load, simply
        returning immediately is sufficient.
        """
        raise NotImplementedError

    def validate_current_page(self):
        """Ensure that the browser is on the correct page before calling a method.

        To be used to make sure methods on this page are not called while the
        browser is on a different page
        """
        try:
            self._validate_url(self._browser.current_url)
            self._wait_until_page_is_loaded()
        except InvalidURLError:
            raise WrongPageForMethodError()

    @abstractmethod
    def _validate_url(self, url):
        """
        Ensure that the given URL is a valid URL for this page type

        :param url: the url to validate
        :type url: str
        """
        raise NotImplementedError

    @classmethod
    def require_user_type(cls, user_type: UserType):
        """
        Throw an error if the browser is not currently logged in as the required user type.

        To be used as a decorator for Page subclass methods that require the
        browser to be logged in as a specific user type.

        :param func: a page method
        :type func: function
        :param user_type: the user type to require
        :type user_type: UserType
        """

        def require_user_type_decorator(func: Callable):
            def inner_func(self, *args, **kwargs):
                if not self._browser.user_type == user_type:
                    raise InvalidUserTypeError("Invalid user type for method")
                return func(self, *args, **kwargs)

            inner_func.__doc__ = func.__doc__
            return inner_func

        return require_user_type_decorator


    @property
    def url(self):
        """Get the page's url"""
        return self._url



from autohandshake.src import HandshakeBrowser
from autohandshake.src.exceptions import InvalidURLError, InvalidPasswordError
from autohandshake.src.Pages.LoginPage import LoginPage
from autohandshake.src.constants import MAX_WAIT_TIME
import keyring


class HandshakeSession:
    """
    A Handshake browsing session.

    Should be used as a context manager. Example:
    ::

        with HandshakeSession(school_url, email) as browser:
            # do something

    """

    def __init__(self, login_url: str, email: str, password: str = None, max_wait_time: int = MAX_WAIT_TIME):
        """
        :param login_url: a valid Handshake homepage url of the form
                          "https://[school].joinhandshake.com"
        :type login_url: str
        :param email: a valid handshake admin email
        :type email: str
        :param password: the password associated with the given email. If
                         password is not specified, HandshakeSession will instead
                         look for a password associated with the given login_url
                         and email within the machine's default password manager
                         (e.g. Windows Credential Manager, macOS Keychain, etc.).
                         **This is the preferred usage**, since it is more secure.
        :type password: str
        :param max_wait_time: the maximum time to wait for something to load
                              before throwing a timeout error
        :type max_wait_time: int
        """
        self._login_url = login_url
        self._email = email
        if password:
            self._password = password
        else:
            self._password = keyring.get_password(self._login_url, self._email)
            if not self._password:
                raise InvalidPasswordError("Keyring unable to find password for "
                                           f"service {login_url} and user {email}")
        self._browser = HandshakeBrowser(max_wait_time=max_wait_time)

    def __enter__(self) -> HandshakeBrowser:
        """
        Open a web browser and log into Handshake, beginning the session

        :returns: a logged-in HandshakeBrowser
        :rtype: HandshakeBrowser
        """
        self._browser.get(self._login_url)
        if self._school_is_invalid():
            raise InvalidURLError('Invalid school in login URL')
        # try to log in via the new style of login page if the old one fails
        login_page = LoginPage(self._login_url, self._browser)
        login_page.login(self._email, self._password)
        self._browser.update_constants()
        return self._browser

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close the web browser"""
        self.close()

    def close(self):
        """Close the Handshake Session.

        To be called if HandshakeSession is *not* used as a context manager.
        """
        self._browser.quit()

    def _school_is_invalid(self) -> bool:
        """
        Determine whether or not the Handshake login URL's school is invalid.

        If the page asks the user to select a school before signing in, that
        means the URL had an invalid school to begin with.

        :return: True if the school is invalid, false otherwise
        :rtype: bool
        """
        return self._browser.element_exists_by_xpath('//span[text()=\'Please '
                                                     'select your school to '
                                                     'sign in.\']')



from typing import Optional

from autohandshake.src.HandshakeBrowser import HandshakeBrowser, UserType
from autohandshake.src.Pages import Page
from autohandshake.src.constants import BASE_URL


class AppointmentTypePage(Page):
    """A settings page for a single appointment type."""

    def __init__(self, type_id: int, browser: HandshakeBrowser):
        """
        :param type_id: the id of the appointment type to load
        :type type_id: int
        :param browser: a logged-in HandshakeBrowser
        :type browser: HandshakeBrowser
        """
        self._id = type_id
        super().__init__(f'{BASE_URL}/appointment_types/{type_id}/edit', browser)

    @Page.require_user_type(UserType.STAFF)
    def get_settings(self) -> dict:
        """
        Get the settings for this page's appointment type.

        The method returns a dictionary of settings values of the form:
        ::

            {
                'id': 238492,
                'name': 'STEM Resume Appointment',
                'description': 'An appointment for STEM students who want to improve their resume.',
                'length': 30,
                'categories': ['Resume/Cover Letter Review'],
                'drop_in_enabled': False,
                'pre_message': 'Please bring your resume with you to the appointment.',
                'pre_survey': None,
                'post_message': 'Thank you for stopping by!',
                'post_survey': 'Resume Review Satisfaction Survey',
                'staff_survey': 'Main Career Center Staff Survey',
                'school_years': ['Sophomore', 'Junior', 'Senior'],
                'cum_gpa_required': True,
                'cum_gpa': 3.0,
                'major_groups': ['major_group: Computer Science', 'major_group: Chemical Engineerng',
                                 'major_group: Biology', 'indv_major: Biological Engineering'],
                'colleges': ['School of Engineering', 'School of Arts & Sciences'],
                'labels': ['appointments allowed'],
                'career_clusters': ['Science Careers', 'Engineering Careers']
            }

        :return: a dictionary of all the appointment type settings fields
        :rtype: dict
        """
        settings = {}

        # collect fields from first page
        settings['id'] = self._id
        settings['name'] = self._browser.get_element_attribute_by_xpath(
            "//*[@id='appointment_type_name']", "value")
        settings['description'] = self._browser.get_element_attribute_by_xpath(
            "//*[@id='appointment_type_description']", "value")
        settings['length'] = self._parse_length()
        settings['categories'] = self._parse_multi_option_appt_type_field(
            's2id_appointment_type_appointment_category_ids')
        settings['drop_in_enabled'] = self._browser.element_is_selected_by_xpath(
            "//*[@id='appointment_type_drop_in_enabled']")
        settings['pre_message'] = self._browser.get_element_attribute_by_xpath(
            "//*[@id='appointment_type_pre_message']", "value")
        settings['pre_survey'] = self._parse_survey_drop_down('s2id_appointment_type_pre_survey_id')
        settings['post_message'] = self._browser.get_element_attribute_by_xpath(
            "//*[@id='appointment_type_post_message']", "value")
        settings['post_survey'] = self._parse_survey_drop_down('s2id_appointment_type_post_survey_id')
        settings['staff_survey'] = self._parse_survey_drop_down('s2id_appointment_type_advisor_survey_id')

        # switch to second page
        self._browser.click_element_by_xpath("//a[text()='Requirements']")
        self._browser.wait_until_element_exists_by_xpath(
            "//*[@id='student_screen_cumulative_gpa_required']")

        # collect fields from second page
        settings['school_years'] = self._parse_school_year_requirements()
        settings['cum_gpa_required'] = self._browser.element_is_selected_by_xpath(
            "//*[@id='student_screen_cumulative_gpa_required']")
        settings['cum_gpa'] = self._parse_gpa()
        settings['major_groups'] = self._parse_major_group_qualifications()
        settings['colleges'] = self._parse_multi_option_appt_type_field(
            's2id_student_screen_college_ids')
        settings['labels'] = self._parse_multi_option_appt_type_field(
            's2id_student_screen_institution_label_ids')
        settings['career_clusters'] = self._parse_multi_option_appt_type_field(
            's2id_student_screen_custom_job_interest_option_ids')

        return settings

    def _validate_url(self, url):
        """
        Ensure that the given URL is a valid URL.

        Since the URL is not entered by the user, it is always valid.

        :param url: the url to validate
        :type url: str
        """
        return

    def _wait_until_page_is_loaded(self):
        """Wait until the page has finished loading."""
        self._browser.wait_until_element_exists_by_xpath("//input[@id='appointment_type_name']")

    def _parse_multi_option_appt_type_field(self, parent_div_id: str) -> Optional[list]:
        """
        Extract all selected values from a multi-value text field.

        :param parent_div_id: the id of the parent div of the field
        :type parent_div_id: str
        :return: a list of all selected values in the field
        :rtype: list
        """
        values = self._browser.get_elements_attribute_by_xpath(
            f"//*[@id='{parent_div_id}']//*[@class='select2-search-choice']//div",
            "text")
        if values:
            return values
        return None

    def _parse_survey_drop_down(self, parent_div_id: str) -> Optional[str]:
        """
        Extract the name of the selected survey from a survey drop-down.

        :param parent_div_id: the id of the parent div of the drop-down
        :type parent_div_id: str
        :return: the name of the selected survey, if any
        :rtype: str
        """
        non_choices = ['Pre Survey...', 'Post Survey...', 'Advisor Survey...']
        survey_name = self._browser.get_element_attribute_by_xpath(
            f"//*[@id='{parent_div_id}']//*[@class='select2-chosen']", "text")
        if survey_name in non_choices:
            return None
        return survey_name

    def _parse_school_year_requirements(self) -> Optional[list]:
        """
        Extract the selected school years from the requirements page

        :return: a list of selected school years, if any
        :rtype: list
        """
        element_mapping = {
            1: 'Freshman',
            2: 'Sophomore',
            3: 'Junior',
            4: 'Senior',
            5: 'Masters',
            6: 'Doctorate',
            7: 'Alumni',
            8: 'Postdoctoral Studies',
        }
        school_years_required = []
        for i in range(1, 9):
            if self._browser.element_is_selected_by_xpath(
                    f"//*[@id='student_screen_school_year_ids_{i}']"):
                school_years_required.append(element_mapping[i])
        if not school_years_required:
            return None
        return school_years_required

    def _parse_major_group_qualifications(self) -> Optional[list]:
        """
        Extract the selected major groups and majors from the requirements page.

        Selected major groups will be prefixed with "major_group: " and any
        individual majors will be prefixed with "indv_major: "

        :return: a list of selected majors and major groups, if any
        :rtype: list
        """
        major_qualifications = []
        groups = self._browser.get_elements_attribute_by_xpath("//*[contains(@class, 'label-primary')]/span", "text")
        indv_majors = self._browser.get_elements_attribute_by_xpath("//*[contains(@class, 'label-info')]/span", "text")
        for group in groups:
            major_qualifications.append('major_group: ' + group)
        for indv_major in indv_majors:
            major_qualifications.append('indv_major: ' + indv_major)
        if not major_qualifications:
            return None
        return major_qualifications

    def _parse_gpa(self) -> Optional[float]:
        """Get the required GPA of a type, if any"""
        gpa = self._browser.get_element_attribute_by_xpath(
            "//*[@id='student_screen_cumulative_gpa']", "value")
        if gpa:
            return float(gpa)
        return None

    def _parse_length(self) -> Optional[int]:
        length = self._browser.get_element_attribute_by_xpath(
            "//*[@id='appointment_type_length']", "value")
        if length:
            return int(length)
        return None
